/* AVIJOIN.CPP 1.0.0 */
/* Written by Samuele Zannoli */
// This program is put in the public domain

#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <mmsystem.h>
#include <vfw.h>

// Let's tell the linker we need 2 libraries
#pragma comment(lib, "winmm.lib")
#pragma comment(lib, "vfw32.lib")

// The 2 avi files (source and destination)
PAVIFILE pfile;
PAVIFILE pfiledst;
// Video stream
PAVISTREAM pstream;
PAVISTREAM pstreamdst;
// Audio stream
PAVISTREAM pstreamwav;
PAVISTREAM pstreamwavdst;
// Information about the streams
AVISTREAMINFO streaminfo;
AVISTREAMINFO streaminfowav;
// True if the source avi file has audio too
BOOL hasaudio;
// Number of streams
int numstreams;
// Number of frames in the source file
int numframes;
// Number of the first frame in the source file
int firstframe;
// X and Y dimension of the frames (in pixels)
int dimx,dimy;
// Frames per second
float framerate;
// Audio samples per second (if the audio was digitized at 11050 Hz
// then samplefreq=11050 )
float samplefreq;
// Number of audio samples per frame (equal to samplefreq/framerate)
float samplexframe;
// Buffer used to transfer data
BYTE buffer[32768];

int opensrc(LPCTSTR elem);
int closesrc(void);
int opendst(LPCTSTR elem);
int closedst(void);
LONG copyframes(LONG lStart,LONG lDest);

int main(int argc,char *argv[])
{
LONG n,nn,v;
int nv,i,f;
char dest[256],src[256];

	printf("AVIJOIN 1.0.0 - Join AVI files into a single one\n\n");
	f=2;
	if (argc < 3)
	{
		printf("Number of AVI files to join: ");
		scanf("%d",&nv);
		f=0;
	}
	if (argc > 1)
	{
		strcpy(dest,argv[1]);
	}
	else
	{
		printf("Destination file: ");
		scanf("%s",dest);
	}
	nn=0;
	// Let's initialize the AVI file library
	AVIFileInit();
	for (i=1; (i <= nv) || (f > 0) ;i++)
	{
		if (f == 0)
		{
			printf("Source file %d: ",i);
			scanf("%s",src);
		}
		else
		{
			strcpy(src,argv[f]);
		}
		// Open one of the avi files to join
		if (opensrc(src))
		{
			if (i > 1)
				closedst();
			AVIFileExit();
			exit(1);
		}
		// Open the 'joined' avi file
		if (i == 1)
			if (opendst(dest))
			{
				closesrc();
				AVIFileExit();
				exit(1);
			}
		printf("Frames processed in %s:\n",src);
		n=firstframe;
		for (;;)
		{
			// Copy a frame from source to destination
			v=copyframes(n,nn);
			if (v == 0)
				break;
			n=n+v;
			nn=nn+v;
			if (!(nn % 50))
				printf("\t This file: %05d Total: %05d\r",n,nn);
		}
		// Close the source avi file
		closesrc();
		printf("\t This file: %05d Total: %05d\n",n,nn);
		if (f != 0)
		{
			f++;
			if (f == argc)
				break;
		}
	}
	// Close the joined avi file
	closedst();
	// We don't need the AVI file library any more
	AVIFileExit();
	return 0;
}

// OPENSRC: Opens an AVI file with its video and audio streams and
// initializes some global variables. ELEM is the name of the avi file
int opensrc(LPCTSTR elem)
{
AVIFILEINFO aviinfo;

	// Open the avi file, retur 1 if error
	if (AVIFileOpen(&pfile,elem,OF_READ,NULL))
		return 1;
	// Get info about the file
	AVIFileInfo(pfile,&aviinfo,sizeof(AVIFILEINFO));
	numstreams=aviinfo.dwStreams;
	// Open the first video stream (number 0)
	AVIFileGetStream(pfile,&pstream,streamtypeVIDEO,0);
	// Get info about the video stream
	AVIStreamInfo(pstream,&streaminfo,sizeof(AVISTREAMINFO));
	dimx=streaminfo.rcFrame.right;
	dimy=streaminfo.rcFrame.bottom;
	firstframe=streaminfo.dwStart;
	numframes=streaminfo.dwLength-firstframe;
	// Compute the frame rate
	framerate=(float)streaminfo.dwRate/(float)streaminfo.dwScale;
	hasaudio=FALSE;
	if (numstreams > 1)
	{
		// Open the audio stream
		if (AVIFileGetStream(pfile,&pstreamwav,streamtypeAUDIO,0))
			return 0;
		hasaudio=TRUE;
		// Get info about the audio stream
		AVIStreamInfo(pstreamwav,&streaminfowav,sizeof(AVISTREAMINFO));
		samplefreq=(float)streaminfowav.dwRate/(float)streaminfowav.dwScale;
		samplexframe=samplefreq/framerate;
	}
	return 0;
}

// CLOSESRC: Closes the source avi file
int closesrc(void)
{
	// Close the video stream
	AVIStreamRelease(pstream);
	if (hasaudio)
		// Close the audio stream
		AVIStreamRelease(pstreamwav);
	// Close the avi file
	AVIFileRelease(pfile);
	return 0;
}

// OPENDST: Creates the 'joined' avi file and two streams (one for video and
// one for audio) in it. ELEM is the name of the avi file
int opendst(LPCTSTR elem)
{
BITMAPINFOHEADER bi;
WAVEFORMATEX wf;
LONG lStreamSize;

	// Create the avi file
	if (AVIFileOpen(&pfiledst, elem, OF_WRITE | OF_CREATE, NULL))
		return 1;
	// Create a video stream with the same characteristics of the video stream
	// in the source avi file (the video stream opened with the last call 
	// to OPENSRC)
	AVIFileCreateStream(pfiledst, &pstreamdst, &streaminfo);
	lStreamSize = sizeof(bi);
	// Set the format of the video stream (if we forget to do this the avi file
	// library won't write any data in the stream)
    AVIStreamReadFormat(pstream, 0, &bi, &lStreamSize);
	AVIStreamSetFormat(pstreamdst, 0, &bi, sizeof(bi));
	if (hasaudio)
	{
		// Same as above for the audio stream
		AVIFileCreateStream(pfiledst, &pstreamwavdst, &streaminfowav);
		lStreamSize = sizeof(wf); 
		AVIStreamReadFormat(pstreamwav, 0, &wf, &lStreamSize);
		AVIStreamSetFormat(pstreamwavdst, 0, &wf, sizeof(wf));
	}
	return 0;
}

// CLOSEDST: Close the joined avi file
int closedst(void)
{
	// Close the video stream
	AVIStreamRelease(pstreamdst); 
	if (hasaudio)
	{
		// Close the audio stream
		AVIStreamRelease(pstreamwavdst);
	}
	AVIFileRelease(pfiledst); 
	return 0;
}

// COPYFRAMES: Copy one video frame and its audio from the source file to
// destination. LSTART is the frame number in the source file, LDEST is
// where the frame is written in the destination file
LONG copyframes(LONG lStart,LONG lDest)
{
// Bytes read and written from a stream
LONG plBytes,bw;
// Samples read and written from a stream
LONG plSamples,sw,ps;
// Audio samples to read
LONG wavsamples,fsam;
static LONG lastfsam,lastsam;
DWORD flags;
int err;

	// Read 1 frame at position lStart from the source video stream, put it in
	// the buffer (that's 32768 bytes big), and tell me how many bytes and how
	// many samples you have actually read
	AVIStreamRead(pstream, lStart, 1, 
		          buffer, 32768, &plBytes, &plSamples); 
	// If we read 0 frames return 0
	if (plSamples == 0)
		return 0;
	flags=0;
	// Is this a key frame ?
	if (AVIStreamIsKeyFrame(pstream, lStart))
		flags=AVIIF_KEYFRAME;
	// Write the frame to the destination video stream
	err=AVIStreamWrite(pstreamdst, lDest, plSamples, 
				   buffer, plBytes, flags, &sw, &bw);
	ps=plSamples;
	if (hasaudio)
	{
		flags=0;
		if (lDest == 0)
			lastsam=lastfsam=0;
		if (lStart == firstframe)
			lastfsam=lastsam;
		// Compute the number of audio samples to copy
		wavsamples=(long)(samplexframe*(float)(lStart+1-firstframe));
		wavsamples=wavsamples-(long)(samplexframe*(float)(lStart-firstframe));
		// Read wavsamples audio samples at position lStart*samplexframe
		fsam=(long)((lStart-firstframe)*samplexframe);
		AVIStreamRead(pstreamwav, fsam, wavsamples, 
					  buffer, 32768, &plBytes, &plSamples); 
		// Write the audio samples
		lastsam=lastfsam+fsam+plSamples;
		err=AVIStreamWrite(pstreamwavdst, lastfsam+fsam, plSamples, 
						   buffer, plBytes, flags, &sw, &bw);
	}

	return ps;
}
